from ._internal import DEFAULT, defaultarguments, DoreahConfig

import os
import yaml

config = DoreahConfig("localisation",
	folder="localisation",
	default="en_GB"
)


locs = {}

for f in os.listdir(config["folder"]):
	if f.endswith(".yml"):
		with open (os.path.join(config["folder"],f)) as locfile:
			data = yaml.safe_load(locfile)
			locs[f.split(".")[0]] = data

print(locs)

localize = {
	lang:lambda x,lang=lang:_localize(x,lang) for lang in locs
}

def _localize(key,lang):
	if key in locs[lang]["strings"]:
		return locs[lang]["strings"][key]
	elif "inherit" in locs[lang]:
		return _localize(key,locs[lang]["inherit"])
	else:
		return locs[config["default"]]

class Newstr(str):
	def localize(self,lang,**keys):
		return _localize(str,lang).format(**keys)

__builtins__.__dict__["str"] = Newstr



from ._internal import DEFAULT, defaultarguments, DoreahConfig
import os
import shutil

config = DoreahConfig("filesystem",
)


def relative_file_list(pth,exclude_dotfiles=True):
	for (dirpath,dirnames,filenames) in os.walk(pth):
		reldir = os.path.relpath(dirpath,start=pth)
		for f in filenames:
			if f.startswith(".") and exclude_dotfiles: continue
			relfilepath = os.path.join(reldir, f)
			if relfilepath.startswith("./"): relfilepath = relfilepath[2:]
			yield relfilepath



class FileWrapper:
	def __init__(self,pth):
		self.pth = pth

	def open(self,mode="r"):
		return open(self.pth,mode)

	def mtime(self):
		return int(os.path.getmtime(self.pth))

	def ext(self):
		return self.pth.split(".")[-1].lower()

	def remove(self):
		os.remove(self.pth)

###
# These don't technically represent relative paths, but absolute paths with
# context information about what they're relative to
###

class RelativePath:
	def __init__(self,relpath,root=None):
		if root is None: root = os.curdir
		self.pth = os.path.join(root,relpath)
		self.root = root
		self.relpath = relpath

	def __str__(self):
		return self.relpath

	def parentview(self):
		newroot, parent = os.path.split(self.root)
		newrel = os.path.join(parent, self.relpath)
		return self.__class__(relpath=newrel,root=newroot)

	def copyinto(self,targetdir):
		target = os.path.join(targetdir.fullpath,self.relpath)
		os.makedirs(os.path.dirname(target),exist_ok=True)
		shutil.copytree(self.pth,target)

class RelativeFile(RelativePath,FileWrapper):
	def copyinto(self,targetdir,rename=None):
		target = os.path.join(targetdir.fullpath,self.relpath)
		os.makedirs(os.path.dirname(target),exist_ok=True)
		if rename is not None: target = os.path.join(os.path.dirname(target),rename)
		shutil.copy(self.pth,target)



def directory_dict(pth,file_values=FileWrapper,exclude_dotfiles=True):
	contents = {}
	for e in os.listdir(pth):
		if e.startswith(".") and exclude_dotfiles: continue
		fullpth = os.path.join(pth,e)
		if os.path.isdir(fullpth):
			contents[e] = directory_dict(fullpth)
		else:
			contents[e] = file_values(fullpth)
	return contents


class Directory:
	def __init__(self,path,exclude_dotfiles=True):
		self.fullpath = path
		self.folderdict = {}
		for e in os.listdir(path):
			if e.startswith(".") and exclude_dotfiles: continue
			fullpth = os.path.join(path,e)
			if os.path.isdir(fullpth):
				self.folderdict[e] = Directory(fullpth)
			else:
				self.folderdict[e] = RelativeFile(e,root=self.fullpath)

	def __getitem__(self,node):
		try:
			return self.folderdict[node]
		except:
			os.makedirs(os.path.join(self.fullpath,node))
			self.folderdict[node] = {}
			return self.folderdict[node]

	def relfile(self,relpath):
		return RelativeFile(relpath=relpath,root=self.fullpath)


	# returns all files in this directory tree relative to it
	def allfiles(self,exclude=()):
		for nodename in self.folderdict:
			if nodename in exclude: continue
			node = self.folderdict[nodename]
			if isinstance(node,Directory):
				#yield from r.allfiles()
				for f in node.allfiles():
					yield f.parentview()
			else:
				yield node



import os
import shutil
import yaml
import random

from ._internal import DEFAULT, defaultarguments, DoreahConfig


config = DoreahConfig("settings",
	files=["settings.ini","settings.conf","configuration.ini","configuration.conf"],
	comment=["#"],
	category=("[","]"),
	onlytext=False,
	environ_prefix=None
)



def _interpret(text):
	if config["onlytext"]: return text

	if text.lower() in ["true","yes"]: return True
	if text.lower() in ["false","no"]: return False
	if text.lower() in ["none","nan","n/a",""]: return None
	if text.startswith("[") and text.endswith("]"):
		list = []
		buffer = ""
		string = None
		stringended = False
		for c in text[1:-1]:
			if stringended and c != ",": pass	#after a string is done, skip to the next delimiter
			elif c == '"' and string is None:
				string = '"'		# start new string
				buffer = '"'
			elif c == "'" and string is None:
				string = "'"		# start new string
				buffer = "'"
			elif c == '"' and string is '"':
				string = None		# terminate string
				stringended = True
				buffer += '"'
			elif c == "'" and string is "'":
				string = None		# terminate string
				stringended = True
				buffer += "'"
			elif c == "," and string is None:
				list.append(buffer.strip())
				buffer = ""
				stringended = False
			else: buffer += c

		list.append(buffer.strip())
		return [_interpret(entry) for entry in list]

	if text.startswith("'") and text.endswith("'"): return text[1:-1]
	if text.startswith('"') and text.endswith('"'): return text[1:-1]
	try:
		return int(text)
	except:
		pass
	try:
		return float(text)
	except:
		pass
	return text


# get settings
# keys			list of requested keys. if present, will return list of according values, if not, will return dict of key-value pairs
# files			which files to parse. later files (higher indicies) will overload earlier ones
# environ_prefix	check environment variables with this prefix first
# prefix		only request keys with a certain prefix. key filter will still be applied if present
# cut_prefix	return keys without the prefix
# category		return only keys of specific category
# raw			do not interpret data type, only return strings
@defaultarguments(config,files="files",environ_prefix="environ_prefix")
def get_settings(*keys,files=DEFAULT,environ_prefix=DEFAULT,prefix="",cut_prefix=False,category=None,raw=False):
	"""Get the active settings value for all supplied keys. Get a dict of all settings (or only filtered) if no keys are supplied.

	:param string keys: Setting keys to be read
	:param list files: List of settings files from least to most authorative
	:param string environ_prefix: Check environment variables with this prefix
	:param string prefix: Only return settings keys beginning with this string
	:param boolean cut_prefix: Drop specified prefix from key names
	:param string category: Only return settings of this category
	:param boolean raw: Do not parse data types, return all values as strings
	:return: Tuple of values for all specified keys. If no keys are specified, dict of all settings that satisfy the constraints."""

	allsettings = {}

	for f in files:
		if not os.path.exists(f): continue

		# if category is specified, ignore all entries by default and switch when we encounter the right heading
		ignore = False if category is None else True

		with open(f) as file:
			lines = file.readlines()


		for l in lines:
			# clean up line
			l = l.replace("\n","")
			for symbol in config["comment"]:
				l = l.split(symbol)[0]
			l = l.strip()

			# check if valid settings entry
			if l == "": continue
			if l.startswith(config["category"][0]):
				# ignore category headers if we don't care
				if category is None: continue
				# otherwise, find out if this is the category we want
				else:
					if l.endswith(config["category"][1]):
						cat = l[len(config["category"][0]):-len(config["category"][1])]
						ignore = not (cat == category) #if this is the right heading, set ignore to false
					continue

			if ignore: continue

			if "=" not in l: continue

			# read
			(key,val) = l.split("=")
			key = key.strip()
			val = val.strip() if raw else _interpret(val.strip())

			if key.startswith(prefix):
				# return keys without the common prefix
				if cut_prefix:
					allsettings[key[len(prefix):]] = val

				# return full keys
				else:
					allsettings[key] = val


	# environment variables
	if environ_prefix is not None:
		v = os.environ
		for key in v:
			if key.startswith(environ_prefix):
				val = _interpret(v[key])
				key = key[len(environ_prefix):]
				allsettings[key] = val

	# no arguments means all settings
	if len(keys) == 0:
		return allsettings

	# specific keys requested
	else:
		if len(keys) == 1: return allsettings.get(keys[0])
		else: return [allsettings.get(k) for k in keys]


def update_settings(file,settings,create_new=False):
	"""Updates all settings in a file with settings from the supplied dictionary.

	:param string file: Settings file to write to
	:param dictionary settings: Settings data to write
	:param boolean create_new: Whether to write settings that are not already present in the file"""

	if not os.path.exists(file): open(file,"w").close()

	with open(file,"r") as origfile:
		lines = origfile.readlines()

	newlines = []

	for origline in lines:
		l = origline
		# clean up line
		l = l.replace("\n","")
		for symbol in config["comment"]:
			l = l.split(symbol)[0]
		l = l.strip()

		# check if valid settings entry
		if l == "":
			newlines.append(origline)
			continue
		if l.startswith(config["category"][0]):
			newlines.append(origline)
			continue
		if "=" not in l:
			newlines.append(origline)
			continue

		# read
		(key,val) = l.split("=")
		key = key.strip()
		val = val.strip()


		if key not in settings:
			newlines.append(origline)
			continue

		else:
			#print("Found key")
			newline = origline.split("=",1)
			#print({"linepart":newline[1],"keytoreplace":val,"new":settings[key]})
			newline[1] = newline[1].replace(val,'"' + settings[key] + '"' if isinstance(settings[key],str) else str(settings[key]),1)
			newline = "=".join(newline)
			newlines.append(newline)

			del settings[key]

	if create_new:
		# settings that were not present in the file
		for key in settings:
			newlines.append(key + " = " + str(settings[key]) + "\n")

	with open(file,"w") as newfile:
		newfile.write("".join(newlines))





def update(source="default_settings.ini",target="settings.ini"):
	"""Updates a user settings file to a newer format from a default file without overwriting user's settings.

	:param string source: Default settings file that provides the format
	:param string target: Local settings file that provides higher authority on settings values"""

	if not os.path.exists(target):
		shutil.copyfile(source,target)
	else:
		usersettings = get_settings(files=[target],raw=True)
		shutil.copyfile(source,target)
		update_settings(target,usersettings)




class YAMLBackedConfig(dict):
	def __init__(self,file="conf.yml",defaults={},reload_every=25):
		self.file = file
		self.reload_every = reload_every
		super().__init__(defaults)
		if os.path.exists(self.file):
			self.loadfromdisk()

		with open(self.file,"w") as df:
			yaml.dump(self.tonormaldict(),df)

	def loadfromdisk(self):
		with open(self.file) as df:
			d = yaml.safe_load(df)
			if d is not None:
				self.update(d)

	def tonormaldict(self):
		return {k:self[k] for k in self}

	def __getitem__(self,key):
		if key not in self or random.choice(range(0,self.reload_every)) == 0:
			self.loadfromdisk()

		return super().__getitem__(key)



import math
import datetime
from threading import Thread
import os
import sys
import pickle

from ._internal import DEFAULT, defaultarguments, DoreahConfig
from .persistence import save, load, delete, size
from .regular import repeathourly
from .logging import log


_caches = {}

config = DoreahConfig("caching",
	maxsize=math.inf,
	maxmemory=math.inf,
	maxstorage=16*1024*1024*1024,
	maxage=60*60*24*1,
	maxage_negative=60*60*24*1,
	lazy_refresh=True,
	folder="cache"
)


class Cache:
	"""Dictionary-like object to store key-value pairs up to a certain amount and discard them after they expire.

	:param integer maxsize: Amount of entries the cache should hold before discarding old entries
	:param integer maxmemory: Soft memory limit for the cache (in bytes), not meant for precision
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param boolean persistent: Whether this cache should be persistent through program restarts
	:param string name: If persistent, this is the filename used"""

	@defaultarguments(config,maxsize="maxsize",maxmemory="maxmemory",maxage="maxage",maxage_negative="maxage_negative")
	def __init__(self,maxsize=DEFAULT,maxmemory=DEFAULT,maxage=DEFAULT,maxage_negative=DEFAULT,persistent=False,name=None):

		self.maxsize = maxsize
		self.maxmemory = maxmemory
		self.maxage = maxage
		self.maxage_negative = maxage_negative
		self.persistent = persistent
		if self.persistent:
			self.name = name
			try:
				obj = load(name,folder=config["folder"])
				if obj is not None:
					self.cache,self.times = obj
					self.changed = False
					self._maintenance()
					return
			except:
				pass
		# if either no object loaded, or not persistent in the first place
		self.cache = {}
		self.times = {}
		self.changed = False
		self._maintenance()

	def __del__(self):
		self._maintenance()

#	def create(name="defaultcache",**kwargs):
#		"""Create a new Cache object, preinitializing it from disk if applicable
#
#		:param string name: Name of the object, consistent between sessions
#		:param kwargs: Standard arguments of :class:`Cache` object creation
#		:return: Newly created object"""
#
#		obj = load(name,folder=config["folder"])
#		if obj is not None: return obj
#
#		obj = Cache(**kwargs,persistent=True,name=name)
#		return obj

	def __contains__(self,key):
		#if key not in self.cache: return False

		#now_stamp = int(datetime.datetime.utcnow().timestamp())
		#cache_stamp = self.times.get(key)

		#if self.cache[key] is None and (now_stamp - cache_stamp) < self.maxage_negative: return True
		#if self.cache[key] is not None and (now_stamp - cache_stamp) < self.maxage: return True

		#return False
		try:
			self.get(key)
			return True
		except KeyError:
			return False


	def get(self,key,allow_expired=False):
		"""Get the value of a key in the cache.

		:param key: Key to be retrieved
		:param boolean allow_expired: If set to True, entries that have already expired will still be returned.
		:return: Value of the requested key in the cache
		:raises KeyError: No valid entry for the key found."""

		if key not in self.cache: raise KeyError()

		value = self.cache[key]

		if allow_expired: return value

		now_stamp = int(datetime.datetime.utcnow().timestamp())
		cache_stamp = self.times.get(key)



		if value is None and (now_stamp - cache_stamp) < self.maxage_negative: return value
		if value is not None and (now_stamp - cache_stamp) < self.maxage: return value

		raise KeyError()

	def add(self,key,value):
		"""Add an entry to the cache.

		:param key: Key to be added
		:param value: Value of this entry"""

#		if len(self.cache) < self.maxsize: #still have space
#			pass
#		else:
#			self.flush() #try to make space
#			if len(self.cache) < self.maxsize: pass # success
#			else:
#				self.
#				raise MemoryError() # still no space


		self.cache[key] = value
		self.times[key] = int(datetime.datetime.utcnow().timestamp())

		self._onupdate()

	def flush(self):
		"""Flush all expired entries from the cache. This is normally done lazily when needed, so this function does not need to be called manually."""

		now_stamp = int(datetime.datetime.utcnow().timestamp())

		for key in list(self.cache.keys()):
			value = self.cache[key]
			cache_stamp = self.times.get(key)

			if value is None and (now_stamp - cache_stamp) > self.maxage_negative:
				del self.cache[key]
				del self.times[key]
			if value is not None and (now_stamp - cache_stamp) > self.maxage:
				del self.cache[key]
				del self.times[key]


	def _onupdate(self):
		self.changed = True

	@repeathourly
	def _maintenance(self):
		if self.changed:
			if len(self.cache) > self.maxsize or self._size() > self.maxmemory:
				#flush anyway expired entries
				self.flush()

			while len(self.cache) > self.maxsize or self._size() > self.maxmemory:
				log("Expiring oldest entry",module="debug")
				#expire oldest entry
				keys = list(self.times.keys())
				keys.sort(key=lambda k:self.times[k])
				delkey = keys[0]
				del self.cache[delkey]
				del self.times[delkey]

			if self.persistent:
				save((self.cache,self.times),self.name,folder=config["folder"])

			self.changed = False



	def _size(self):
		return sys.getsizeof(pickle.dumps(self.cache))



class _DiskReference:
	def __init__(self,filename):
		self.filename = filename

class DeepCache(Cache):
	"""Dictionary-like object to store key-value pairs with the help of hard disk storage up to a certain amount and discard them after they expire.

	:param integer maxmemory: Soft memory limit for the cache (in bytes), not meant for precision
	:param integer maxstorage: Soft disk space limit for the cache (in bytes), not meant for precision
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param string name: Directory name used for storage"""

	@defaultarguments(config,maxmemory="maxmemory",maxstorage="maxstorage",maxage="maxage",maxage_negative="maxage_negative")
	def __init__(self,maxmemory=DEFAULT,maxage=DEFAULT,maxstorage=DEFAULT,maxage_negative=DEFAULT,name="deepcache"):

		self.maxmemory = maxmemory
		self.maxstorage = maxstorage
		self.maxage = maxage
		self.maxage_negative = maxage_negative

		self.name = name

		try:
			obj = load(self._file(),folder=config["folder"])
			self.cache,self.times,self.counter = obj
			self.changed = False
		except:
			self.cache = {}
			self.times = {}
			self.counter = 0

		self.changed = False

		self._maintenance()

	def get(self,key,allow_expired=False):
		"""Get the value of a key in the cache.

		:param key: Key to be retrieved
		:param boolean allow_expired: If set to True, entries that have already expired will still be returned.
		:return: Value of the requested key in the cache
		:raises KeyError: No valid entry for the key found."""

		if key not in self.cache: raise KeyError()

		value = self.cache[key]
		if isinstance(value,_DiskReference):
			value = load(self._file(value.filename),folder=config["folder"])
		if allow_expired: return value

		now_stamp = int(datetime.datetime.utcnow().timestamp())
		cache_stamp = self.times.get(key)



		if value is None and (now_stamp - cache_stamp) < self.maxage_negative: return value
		if value is not None and (now_stamp - cache_stamp) < self.maxage: return value

		raise KeyError()

	def add(self,key,value):
		"""Add an entry to the cache.

		:param key: Key to be added
		:param value: Value of this entry"""

		# remove old file
		if key in self.cache and isinstance(self.cache[key],_DiskReference):
			delete(self._file(self.cache[key].filename),folder=config["folder"])

		self.cache[key] = value
		self.times[key] = int(datetime.datetime.utcnow().timestamp())

		self._onupdate()

	def flush(self):
		"""Flush all expired entries from the cache. This is normally done lazily when needed, so this function does not need to be called manually."""

		now_stamp = int(datetime.datetime.utcnow().timestamp())

		for key in list(self.cache.keys()):
			value = self.cache[key]
			cache_stamp = self.times.get(key)

			if value is None and (now_stamp - cache_stamp) > self.maxage_negative:
				if isinstance(self.cache[key],_DiskReference): delete(self._file(self.cache[key].filename),folder=config["folder"])
				del self.cache[key]
				del self.times[key]
			if value is not None and (now_stamp - cache_stamp) > self.maxage:
				if isinstance(self.cache[key],_DiskReference): delete(self._file(self.cache[key].filename),folder=config["folder"])
				del self.cache[key]
				del self.times[key]


	def _onupdate(self):
		self.changed = True

	@repeathourly
	def _maintenance(self):
		if self.changed:
			if self._size() > self.maxmemory:
				#flush anyway expired entries
				self.flush()

			while self._size() > self.maxmemory:
				#serialize biggest entry
				keys = list(self.times.keys())
				keys.sort(key=lambda k:sys.getsizeof(pickle.dumps(self.cache[k])),reverse=True)
				keys = [k for k in keys if not isinstance(self.cache[k],_DiskReference)]
				if sys.getsizeof(pickle.dumps(self.cache[keys[0]])) < (512 * 1024):
					break
					# don't serialize tiny stuff
				try:
					movekey = keys[0]
					self._memorytodisk(movekey)
				except:
					break

			while self._disksize() > self.maxstorage:
				print("Disk size " + str(self._disksize()))
				keys = list(self.times.keys())
				keys.sort(key=lambda k:self.times[k])
				keys = [k for k in keys if isinstance(self.cache[k],_DiskReference)]
				try:
					delkey = keys[0]
					print("Deleting " + delkey)
					delete(self._file(self.cache[delkey].filename),folder=config["folder"])
					del self.cache[delkey]
					del self.times[delkey]
				except:
					break


			save((self.cache,self.times,self.counter),self._file(),folder=config["folder"])

			self.changed = False

	def _memorytodisk(self,key):
		print("Moving " + key + " from memory to disk")
		self.counter += 1
		value = self.cache[key]
		filename = str(self.counter)
		save(value,self._file(filename),folder=config["folder"])
		self.cache[key] = _DiskReference(filename)

	def _disktomemory(self,key):
		print("Moving " + key + " from disk to memory")
		filename = self.cache[key].filename
		value = load("./" + self.name + "/" + filename,folder=config["folder"])
		self.cache[key] = value
		delete(self._file(filename),folder=config["folder"])


	def _size(self):
		return sys.getsizeof(pickle.dumps([self.cache[key] for key in self.cache if not isinstance(self.cache[key],_DiskReference)]))

	def _disksize(self):
		#return sum(os.path.getsize("./" + config["folder"] + "/" + self.name + "/" + f) for f in os.listdir("./" + config["folder"] + "/" + self.name))
		sumsize = 0
		for key in self.cache:
			if isinstance(self.cache[key],_DiskReference):
				sumsize += size(self._file(self.cache[key].filename),folder=config["folder"])
		return sumsize

	# gets a relative filename for keys of this deepcache object
	def _file(self,name=None):
		if name is None: name = "root"
		return os.path.join(self.name,name)


# decorator
@defaultarguments(config,maxsize="maxsize",maxage="maxage",maxage_negative="maxage_negative",lazy_refresh="lazy_refresh")
def cached(maxsize=DEFAULT,maxage=DEFAULT,maxage_negative=DEFAULT,lazy_refresh=DEFAULT):
	"""Method decorator to add a proxy cache to a function without keyword arguments.

	:param integer maxsize: Amount of entries the cache should hold before discarding old entries.
	:param integer maxage: Time in seconds entries are valid for after their last update. Entries older than this value are lazily removed, which means they might still be accessible with the ``allow_expired`` argument of the :meth:`get` method.
	:param integer maxage_negative: Time in seconds entries with the ``None`` value are valid. This is useful for negative caching.
	:param boolean lazy_refresh: If True, expired cache entries will still be returned, but trigger a background refresh. Useful if speed is more important than currentness of data."""


	# decorator for lazy refresh
	if lazy_refresh:
		def cacheddecorator(func):

			def newfunc(*args,cache=Cache(maxsize=maxsize,maxage=maxage,maxage_negative=maxage_negative)):
				try:
					result = cache.get(args,allow_expired=True)
				except:
					result = func(*args)
					cache.add(args,result)
					return result

				if args not in cache: # if we just got an expired result, we want to execute the function and add to cache

					def fillcache(cache,args,func):
						cache.add(args,func(*args))

					t = Thread(target=fillcache,args=(cache,args,func,))
					t.start()

				return result

			return newfunc

	# decorator for non-lazy
	else:
		def cacheddecorator(func):

			def newfunc(*args,cache=Cache(maxsize=maxsize,maxage=maxage,maxage_negative=maxage_negative)):
				try:
					return cache.get(args)
				except:
					result = func(*args)
					cache.add(args,result)
					return result

			return newfunc

	return cacheddecorator
